from math import ceil, log2, trunc

from puid.chars import ValidChars
from puid.chars_error import InvalidChars
from puid.puid_error import TotalRiskError


def bits_for_total_risk(total: 0, risk: 0):
    """
    Entropy bits necessary to produce a `total` `puid`s with given `risk` of repeat

    :param total: int
    :param risk: int
    :return float

    >>> bits_for_total_risk(100_000, 1e12)
    72.08241808752197
    """

    def non_neg_int_or_float(value):
        if isinstance(value, int) and 0 <= value:
            return True
        if isinstance(value, float) and value.is_integer() and 0.0 <= value:
            return True
        return False

    if not non_neg_int_or_float(total) or not non_neg_int_or_float(risk):
        raise TotalRiskError('total and risk must be an non-negative integers')

    if total in [0, 1]:
        return 0

    if risk in [0, 1]:
        return 0

    if total < 1000:
        return log2(total) + log2(total - 1) + log2(risk) - 1
    else:
        return 2 * log2(total) + log2(risk) - 1


def bits_per_char(chars):
    """
    Entropy bits per character for either a predefined Chars enum or a string of characters

    :param chars: Either a Chars enum or a string

    raises CharsError subclass if `chars` is invalid

    >>> bits_per_char(Chars.BASE32)
    5.0

    >>> bits_per_char('dingosky_me')
    3.4594316186372973
    """
    if isinstance(chars, ValidChars):
        return log2(len(chars))
    else:
        raise InvalidChars('chars must be an instance of ValidChars')


def bits_for_len(chars, len):
    """
    Bits necessary for a `puid` of length `len` using characters `chars`

    :param chars: Either a Chars enum or a string
    :param len: Desired length of `puid`

    raises CharsError subclass if `chars` is invalid

    >>> bits_for_len('dingosky', 14)
    42
    """
    return trunc(len * bits_per_char(chars))


def len_for_bits(chars, bits):
    """
    Length necessary for a `puid` of `bits` using characters `chars`

    :param chars: Either a Chars enum or a string
    :param bits: Desired `bits` of `puid`

    raises CharsError subclass if `chars` is invalid

    >>> len_for_bits(Chars.SAFE_ASCII, 97)
    15
    """
    return ceil(bits / bits_per_char(chars))


if __name__ == '__main__':  # pragma: no cover
    import doctest

    doctest.testmod()



def symbol():
    bang = ord("!")
    hash_symbol = ord("#") - 1
    open_paren = ord("(") - 5
    colon = ord(":") - 13
    open_square_bracket = ord("[")
    close_square_bracket = ord("]") - 21
    open_curly_bracket = ord("{") - 24

    def encode(n):
        if n == 0:
            return bang
        if n < 5:
            return n + hash_symbol
        if n < 13:
            return n + open_paren
        if n < 20:
            return n + colon
        if n == 20:
            return open_square_bracket
        if n < 24:
            return n + close_square_bracket
        return n + open_curly_bracket

    return encode



def alphanum():
    upper = ord("A")
    lower = ord("a") - 26
    decimal = ord("0") - 52

    def encoder(n):
        if n < 26:
            return n + upper
        if n < 52:
            return n + lower
        return n + decimal

    return encoder


def alphanum_lower():
    return _alphanum_case(False)


def alphanum_upper():
    return _alphanum_case(True)


def _alphanum_case(upper):
    alpha = ord("A") if upper else ord("a")
    decimal = ord("0") - 26

    def encoder(n):
        return n + (alpha if n < 26 else decimal)

    return encoder



from enum import Enum

from puid.chars_error import InvalidChars, NonUniqueChars, TooFewChars, TooManyChars


def valid_chars(chars):
    """
    Tests whether characters are valid.

    raises A CharsError subclass if characters are not valid.

    >>> valid_chars(Chars.HEX)
    True

    >>> valid_chars('dingosky')
    True
    """
    if isinstance(chars, Chars):
        return True

    if not isinstance(chars, str):
        raise InvalidChars('Characters must be a str')

    min_len = 2
    max_len = 256

    if len(chars) < min_len:
        raise TooFewChars(f'Must have at least {min_len} characters')

    if 256 < len(chars):
        raise TooManyChars(f'Exceeded max of {max_len} characters')

    if len(chars) != len(set(chars)):
        raise NonUniqueChars('Characters are not unique')

    for char in chars:
        if not _valid_char(char):
            raise InvalidChars(f'Invalid character with code: {ord(char)}')

    return True


def _valid_char(char):
    code_point = ord(char)

    if 160 < code_point:
        return True

    if char == '!':
        return True
    if code_point < ord('#'):
        return False
    if char == "'":
        return False
    if char == '\\':
        return False
    if char == '`':
        return False
    if ord('~') < code_point:
        return False

    return True


class Chars(Enum):
    """
    Predefined Characters

    These enums are intended to be passed to the `Puid` class initializer for configuration
    """

    ALPHA = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    ALPHA_LOWER = 'abcdefghijklmnopqrstuvwxyz'
    ALPHA_UPPER = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    ALPHANUM = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
    ALPHANUM_LOWER = 'abcdefghijklmnopqrstuvwxyz0123456789'
    ALPHANUM_UPPER = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    BASE16 = '0123456789ABCDEF'
    BASE32 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'
    BASE32_HEX = '0123456789abcdefghijklmnopqrstuv'
    BASE32_HEX_UPPER = '0123456789ABCDEFGHIJKLMNOPQRSTUV'
    CROCKFORD32 = '0123456789ABCDEFGHJKMNPQRSTVWXYZ'
    DECIMAL = '0123456789'
    HEX = '0123456789abcdef'
    HEX_UPPER = '0123456789ABCDEF'
    SAFE_ASCII = '!#$%&()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_abcdefghijklmnopqrstuvwxyz{|}~'
    SAFE32 = '2346789bdfghjmnpqrtBDFGHJLMNPQRT'
    SAFE64 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'
    SYMBOL = '!#$%&()*+,-./:;<=>?@[]^_{|}~'
    WORD_SAFE32 = '23456789CFGHJMPQRVWXcfghjmpqrvwx'

    def __len__(self):
        return len(self.value)


class ValidChars:
    """Base class for PredefinedChars and CustomChars"""

    def __repr__(self):
        return "{0} -> '{1}'".format(self.name, self.value)

    def __len__(self):
        return len(self.value)

    def __iter__(self):
        return iter(self.value)


class PredefinedChars(ValidChars):
    """
    Class for Predefined Chars

    raises InvalidChars if initialized with anything other than a predefined Chars enum

    >>> hex_upper = PredefinedChars(Chars.HEX_UPPER)

    This class is intended for internal use
    """

    def __init__(self, chars):
        """
        Create a PredefinedChars for Chars enum

        :param chars: Chars enum
        """
        if isinstance(chars, Chars):
            self.name = chars.name
            self.value = chars.value
        else:
            raise InvalidChars('PredefinedChars only accepts members of the Chars enum')


class CustomChars(ValidChars):
    """
    Class for Custom Chars

    raises CharsError if initialized with an string of characters that are invalid
    raises InvalidChars if initialized with a predefined Chars enum

    >>> dingosky = CustomChars('dingosky')

    This class is intended for internal use
    """

    def __init__(self, chars):
        """
        Create a CustomChars for a string of characters

        :param chars: A valid string of characters
        """
        if isinstance(chars, Chars):
            raise InvalidChars('Use class PredefinedChars for members of the Chars enum')
        else:
            valid_chars(chars)
            self.name = 'Custom'
            self.value = chars


if __name__ == '__main__':  # pragma: no cover
    import doctest

    doctest.testmod()



from puid.chars import Chars
from puid.encoders.alpha import alpha
from puid.encoders.alpha import alpha_lower
from puid.encoders.alpha import alpha_upper
from puid.encoders.alphanum import alphanum
from puid.encoders.alphanum import alphanum_lower
from puid.encoders.alphanum import alphanum_upper
from puid.encoders.base16 import base16
from puid.encoders.base32 import base32
from puid.encoders.base32 import base32_hex
from puid.encoders.base32 import base32_hex_upper
from puid.encoders.crockford32 import crockford32
from puid.encoders.custom import custom
from puid.encoders.decimal import decimal
from puid.encoders.hex import hex_lower
from puid.encoders.hex import hex_upper
from puid.encoders.safe32 import safe32
from puid.encoders.safe64 import safe64
from puid.encoders.safe_ascii import safe_ascii
from puid.encoders.symbol import symbol
from puid.encoders.word_safe32 import word_safe32


def encoder(chars: Chars):
    if chars.name == Chars.ALPHA.name:
        return alpha()

    if chars.name == Chars.ALPHA_LOWER.name:
        return alpha_lower()

    if chars.name == Chars.ALPHA_UPPER.name:
        return alpha_upper()

    if chars.name == Chars.ALPHANUM.name:
        return alphanum()

    if chars.name == Chars.ALPHANUM_LOWER.name:
        return alphanum_lower()

    if chars.name == Chars.ALPHANUM_UPPER.name:
        return alphanum_upper()

    if chars.name == Chars.BASE16.name:
        return base16()

    if chars.name == Chars.BASE32.name:
        return base32()

    if chars.name == Chars.BASE32_HEX.name:
        return base32_hex()

    if chars.name == Chars.BASE32_HEX_UPPER.name:
        return base32_hex_upper()

    if chars.name == Chars.CROCKFORD32.name:
        return crockford32()

    if chars.name == Chars.DECIMAL.name:
        return decimal()

    if chars.name == Chars.HEX.name:
        return hex_lower()

    if chars.name == Chars.HEX_UPPER.name:
        return hex_upper()

    if chars.name == Chars.SAFE32.name:
        return safe32()

    if chars.name == Chars.SAFE64.name:
        return safe64()

    if chars.name == Chars.SAFE_ASCII.name:
        return safe_ascii()

    if chars.name == Chars.SYMBOL.name:
        return symbol()

    if chars.name == Chars.WORD_SAFE32.name:
        return word_safe32()

    return custom(chars)



# n: 0123456789 01234567 89 01 23456 78901
# c: 0123456789 ABCDEFGH JK MN PQRST VWXYZ

def crockford32():
    zero = ord("0")
    A = ord("A")
    J = ord("J")
    M = ord("M")
    P = ord("P")
    V = ord("V")

    def encoder(n):
        if (n < 10):
            return n + zero
        if (n < 18):
            return n - 10 + A
        if (n < 20):
            return n - 18 + J
        if (n < 22):
            return n - 20 + M
        if (n < 27):
            return n - 22 + P
        return n - 27 + V

    return encoder



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import io
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext
from pathlib import Path

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(join(dirname(__file__), *names), encoding=kwargs.get('encoding', 'utf8')) as fh:
        return fh.read()


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='puid-py',
    version='1.2.0',
    license='MIT',
    description='Simple, flexible and efficient generation of probably unique identifiers (`puid`, '
    'aka random strings) of intuitively specified entropy using pre-defined or custom characters, '
    'including unicode',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Paul Rogers',
    author_email='paul@dingosky.com',
    url='https://github.com/puid/python',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Utilities',
    ],
    project_urls={
        'Documentation': 'https://puid-py.readthedocs.io/',
        'Changelog': 'https://puid-py.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/puid/python/issues',
    },
    keywords=['random string', 'random ID', 'uuid', 'secure token', 'security', 'token'],
    python_requires='>=3.7',
    install_requires=[
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    setup_requires=[
        'pytest-runner',
    ],
    entry_points={
        'console_scripts': [
            'puid = puid.cli:main',
        ]
    },
)
